{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Télécharger cet IPython-Notebook : [Courbe_Elliptique.ipynb](Courbe_Elliptique.ipynb)\n",
    "<p style=\"text-align:center;font-size:200%\">Courbe elliptique</p>\n",
    "****"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "https://www.bibmath.net/forums/viewtopic.php?id=14524"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "On considère la courbe elliptique $\\mathcal{C}$ d'équation de Weierstrass $y^2 = x^3+ax+b$ dans $\\mathbb{Z}/p\\mathbb{Z}$, $p$ premier.\n",
    "\n",
    "On considère l'ensemble $\\mathcal{E}$ des points de la courbe :\n",
    "$$ \\mathcal{E} = \\left\\{ [x, y] \\in \\mathbb{Z}/p\\mathbb{Z}\\times\\mathbb{Z}/p\\mathbb{Z} ~~|~~ y^2 = x^3+ax+b\\right\\} \\cup \\{0\\}$$\n",
    "où $0$ est le point à l'infini de $\\mathcal{C}$.\n",
    "\n",
    "On définit une addition sur $\\mathcal{E}$ qui fait de $\\mathcal{E}$ un groupe abélien, d'élément neutre $0$."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Addition des points"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "D'après le petit théorème de Fermat $n^{p-1}\\equiv 1 \\mod{p}$ si $p$ est premier.  \n",
    "Donc $n^{p-2}\\equiv n^{-1} \\mod{p}$. On suppose naturellement $n\\neq 0$  \n",
    "La fonction Python de calcul de la puissance modulaire `pow` est très rapide."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "def inverse_mod_p(n, p):\n",
    "    \"\"\"Retourne l'inverse de n modulo p premier.\n",
    "    \"\"\"\n",
    "    return pow(n,p-2,p)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "On encapsule les paramètres de la courbe elliptique dans un objet. On définit trois méthodes :\n",
    "* `has_point` qui teste si un point est sur la courbe,\n",
    "* `opp` qui renvoie l'opposé d'un point (c'est le symétrique par rapport à l'axe des abscisses),\n",
    "* `addP` qui renvoie le point correspondant à la somme de deux points."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "class EllipticCurve(object):\n",
    "    \"\"\"Objet représentant la courbe elliptique d'équation y**2 = x**3 + a*x + b\n",
    "       sur le corps Z/pZ (p premier).\n",
    "    \"\"\"\n",
    "    def __init__(self, a, b, p):  \n",
    "        delta = (4*a**3 + 27*b**2)%p  # discriminant\n",
    "        if delta == 0:\n",
    "            raise Exception (\"Discriminant nul : courbe elliptique singulière.\")\n",
    "        self.a = a%p\n",
    "        self.b = b%p\n",
    "        self.p = p\n",
    "        \n",
    "    def __str__(self):\n",
    "        return 'y^2 = x^3 + {}x + {} modulo {}'.format(self.a, self.b, self.p)\n",
    "    \n",
    "    def has_point(self, p1):\n",
    "        if p1 == 0:\n",
    "            return True\n",
    "        return (p1[1] ** 2) % self.p == (p1[0] ** 3 + self.a * p1[0] + self.b) % self.p\n",
    "    \n",
    "    def opp(self, p1):\n",
    "        if p1 == 0:\n",
    "            return 0\n",
    "        return [p1[0], -p1[1]%self.p]\n",
    "    \n",
    "    def addP(self, p1, p2):\n",
    "        if p1 != 0 and self.has_point(p1) == False:\n",
    "            raise Exception (\"Le point {} n'est pas sur la courbe.\".format(p1))\n",
    "        if p2 != 0 and self.has_point(p2) == False:\n",
    "            raise Exception (\"Le point {} n'est pas sur la courbe.\".format(p2))\n",
    "        if p1 == 0:\n",
    "            return p2       # 0 + p2 = p2\n",
    "        if p2 == 0:\n",
    "            return p1       # p1 + 0 = p1\n",
    "        if (p1[0] - p2[0])%self.p == 0 and (p1[1] + p2[1])%self.p == 0 : # si p1 == -p2 mod p\n",
    "            return 0\n",
    "        if (p1[0] - p2[0])%self.p == 0 and (p1[1] - p2[1])%self.p == 0:  # si p1 == p2 mod p\n",
    "            m = (3*p1[0]*p1[0] + self.a)*inverse_mod_p(2*p1[1], self.p) # pente tangente\n",
    "        else:\n",
    "            m = (p2[1] - p1[1])*inverse_mod_p(p2[0] - p1[0], self.p)    # pente sécante\n",
    "        x = (m*m - p1[0] - p2[0])%self.p\n",
    "        return [x, (m*(p1[0] - x) - p1[1])%self.p ]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Application"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [],
   "source": [
    "ec = EllipticCurve(100, 100, 1009)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "y^2 = x^3 + 100x + 100 modulo 1009\n"
     ]
    }
   ],
   "source": [
    "print(ec)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [],
   "source": [
    "p = [12, 1]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "True"
      ]
     },
     "execution_count": 6,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "ec.has_point(p)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Donc le point `p` est sur la courbe.  \n",
    "On demande de calculer `17p` (17 est la clé privée de Bob)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "[102, 275]"
      ]
     },
     "execution_count": 7,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p2 = ec.addP(p, p)      # p2 = 2p\n",
    "p2"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "[401, 373]"
      ]
     },
     "execution_count": 8,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p4 = ec.addP(p2, p2)    # p4 = 4p\n",
    "p4"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "[227, 480]"
      ]
     },
     "execution_count": 9,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p8 = ec.addP(p4, p4)    # p8 = 8p\n",
    "p8"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "[308, 156]"
      ]
     },
     "execution_count": 10,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p16 =ec.addP(p8, p8)    # p16 = 16p\n",
    "p16"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "[237, 355]"
      ]
     },
     "execution_count": 11,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "p17 = ec.addP(p16, p)   # p17 = 17p\n",
    "p17"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "`17p` est la clé publique que Bob envoie à Alice."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Tables d'addition"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Si le module est petit, le nombre de points de la courbe est petit aussi et on peut dresser la table d'addition des points de la courbe elliptique. On n'affiche pas les marges, car la marge gauche est identique à la première colonne et la marge supérieure est identique à la première ligne."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "y^2 = x^3 + 3x + 8 modulo 13\n"
     ]
    }
   ],
   "source": [
    "ec = EllipticCurve(3, 8, 13)\n",
    "print(ec)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Ensemble des points de la courbe :"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[0, [1, 5], [1, 8], [2, 3], [2, 10], [9, 6], [9, 7], [12, 2], [12, 11]]\n"
     ]
    }
   ],
   "source": [
    "e = []\n",
    "for a in range(13):\n",
    "    for b in range(13):\n",
    "        if ec.has_point([a, b]):\n",
    "            if [a, b] not in e:\n",
    "                e.append([a, b])\n",
    "e = [0] + e\n",
    "print(e)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Table d'addition (sans les marges) :"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "    0      [1, 5]   [1, 8]   [2, 3]  [2, 10]   [9, 6]   [9, 7]  [12, 2] [12, 11] \n",
      "  [1, 5]  [2, 10]     0      [1, 8]   [9, 7]   [2, 3]  [12, 2] [12, 11]   [9, 6] \n",
      "  [1, 8]     0      [2, 3]   [9, 6]   [1, 5] [12, 11]  [2, 10]   [9, 7]  [12, 2] \n",
      "  [2, 3]   [1, 8]   [9, 6] [12, 11]     0     [12, 2]   [1, 5]  [2, 10]   [9, 7] \n",
      " [2, 10]   [9, 7]   [1, 5]     0     [12, 2]   [1, 8] [12, 11]   [9, 6]   [2, 3] \n",
      "  [9, 6]   [2, 3] [12, 11]  [12, 2]   [1, 8]   [9, 7]     0      [1, 5]  [2, 10] \n",
      "  [9, 7]  [12, 2]  [2, 10]   [1, 5] [12, 11]     0      [9, 6]   [2, 3]   [1, 8] \n",
      " [12, 2] [12, 11]   [9, 7]  [2, 10]   [9, 6]   [1, 5]   [2, 3]   [1, 8]     0    \n",
      "[12, 11]   [9, 6]  [12, 2]   [9, 7]   [2, 3]  [2, 10]   [1, 8]     0      [1, 5] \n"
     ]
    }
   ],
   "source": [
    "for x in e:\n",
    "    for y in e:\n",
    "        z = ec.addP(x, y)\n",
    "        if z == 0:\n",
    "            print('{:>8}'.format('0   '), end=' ' )\n",
    "        else:\n",
    "            print('{:>8}'.format(str(z)), end=' ' )\n",
    "    print()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Autre courbe elliptique :"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "y^2 = x^3 + 12x + 0 modulo 13\n"
     ]
    }
   ],
   "source": [
    "ec = EllipticCurve(-1, 0, 13)\n",
    "print(ec)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Ensemble des points :"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 16,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "[0, [0, 0], [1, 0], [5, 4], [5, 9], [8, 6], [8, 7], [12, 0]]\n"
     ]
    }
   ],
   "source": [
    "e = []\n",
    "for a in range(13):\n",
    "    for b in range(13):\n",
    "        if ec.has_point([a, b]):\n",
    "            if [a, b] not in e:\n",
    "                e.append([a, b])\n",
    "e = [0] + e\n",
    "print(e)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Table d'addition (sans les marges) :"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 17,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "    0      [0, 0]   [1, 0]   [5, 4]   [5, 9]   [8, 6]   [8, 7]  [12, 0] \n",
      "  [0, 0]     0     [12, 0]   [5, 9]   [5, 4]   [8, 7]   [8, 6]   [1, 0] \n",
      "  [1, 0]  [12, 0]     0      [8, 6]   [8, 7]   [5, 4]   [5, 9]   [0, 0] \n",
      "  [5, 4]   [5, 9]   [8, 6]   [0, 0]     0     [12, 0]   [1, 0]   [8, 7] \n",
      "  [5, 9]   [5, 4]   [8, 7]     0      [0, 0]   [1, 0]  [12, 0]   [8, 6] \n",
      "  [8, 6]   [8, 7]   [5, 4]  [12, 0]   [1, 0]   [0, 0]     0      [5, 9] \n",
      "  [8, 7]   [8, 6]   [5, 9]   [1, 0]  [12, 0]     0      [0, 0]   [5, 4] \n",
      " [12, 0]   [1, 0]   [0, 0]   [8, 7]   [8, 6]   [5, 9]   [5, 4]     0    \n"
     ]
    }
   ],
   "source": [
    "for x in e:\n",
    "    for y in e:\n",
    "        z = ec.addP(x, y)\n",
    "        if z == 0:\n",
    "            print('{:>8}'.format('0   '), end=' ' )\n",
    "        else:\n",
    "            print('{:>8}'.format(str(z)), end=' ' )\n",
    "    print()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "On remarque les trois points [0, 0], [1,0] et [12, 0] sont \"à tangentes verticales\" (si on peut dire !)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "****\n",
    "Le 5/12/2021 - Contact : <A HREF=\"mailto:Rossignol@bribes.org\">Rossignol@bribes.org</A>     "
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.6.4"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
